import fetch from 'node-fetch'
import { URL } from 'url'
import { fileCache } from './files.js'
import { config } from './config.js'
import { hash, viteManifestToDeps } from './util.js'
import { logger } from './logger.js'

export const loadViteManifests = (() => {
  let cachePromise
  let lastCacheTime
  let lastHash

  async function reload () {
    await config.load()
    // vite manifests contains a set of objects with the vite-manifests
    // from the corresponding registered services
    const viteManifests = await Promise.all(config.urls.map(async baseUrl => {
      // fetch the manifests
      const result = await fetch(new URL('/manifest.json', baseUrl))
      if (!result.ok) throw new Error(`Failed to load manifest for url ${result.url} (Status: ${result.status}: ${result.statusText})`)
      try {
        const manifest = await result.json()
        for (const file in manifest) {
          logger.debug(`retrieved ${file} from ${baseUrl}`)
          manifest[file].meta = manifest[file].meta || {}
          manifest[file].meta.baseUrl = baseUrl
        }
        return manifest
      } catch (err) {
        throw new Error(`Failed to load manifest for url ${result.url}: ${err}`)
      }
    }))

    // combine all manifests by keys. With duplicates, last wins
    const viteManifest = viteManifests.reduce((memo, manifest) => Object.assign(memo, manifest), {})
    Object.defineProperty(viteManifest, '__hash__', {
      enumerable: false,
      writable: true
    })

    try {
      viteManifest.__hash__ = hash(viteManifests)
    } catch (err) {
      logger.error(`Failed to calculate hash: ${err.message}`)
    }

    return viteManifest
  }

  return function loadViteManifests ({ useCache = true } = {}) {
    const CACHE_TTL = parseInt(process.env.CACHE_TTL)
    const timeElapsed = () => +((+new Date() - lastCacheTime) / 1000).toFixed(2)

    if (!cachePromise || useCache === false || +new Date() > lastCacheTime + CACHE_TTL) {
      cachePromise = reload()
      cachePromise.then(manifests => {
        if (useCache) logger.info(`reloaded manifests after ${timeElapsed()} seconds`)
        // update cache promise
        const newHash = manifests.__hash__
        if (newHash !== lastHash) {
          if (lastHash) {
            const deps = viteManifestToDeps(manifests)
            // asynchronously rewarm the cache
            fileCache.warmUp(manifests, deps)
          }

          lastHash = newHash
        }
      })
      lastCacheTime = +new Date()
    }
    return cachePromise
  }
})()

export async function getOxManifests () {
  await loadViteManifests().catch(() => {})
  return fileCache.oxManifests
}

export async function getDependencies () {
  // simply catch the error here. This might happen, when one of the UI containers is temporarily not available
  await loadViteManifests().catch(() => {})
  return fileCache.dependencies
}

export async function getCSSDependenciesFor (file) {
  const allDependencies = await getDependencies()
  const dependencies = allDependencies[file] || []
  return dependencies.filter(dep => /\.css/i.test(dep))
}

export async function getVersion () {
  await loadViteManifests().catch(() => {})
  return fileCache.hash
}
